﻿using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;

namespace PuncherTest
{
	class Program
	{
		private enum IPTypeID : byte { IPID_Conn = 2, IPID_ConnOK = 3 };

		private enum MCMode { MCM_NoMC = 0 }

		private const uint C4NetIOUDP__iVersion = 2;

		private const int ConnPacketLen = 41, ConnOKPacketLen = 25;

		static void Main(string[] args)
		{
			try
			{
				if (args.Length < 1) throw new Exception("Invalid arguments specified");

				if (args[0] == "s") // Server
				{
					if (args.Length != 3) throw new Exception("Invalid arguments specified");

					// Parse local and puncher end point
					var bindEP = TryParseIPEndPoint(args[1], IPAddress.Any);
					if (bindEP == null) throw new Exception("Invalid local address specified");
					var puncherEP = TryParseIPEndPoint(args[2], null);
					if (puncherEP == null) throw new Exception("Invalid puncher address specified");

					ServerMain(bindEP, puncherEP);
				}
				else if (args[0] == "c") // Client
				{
					if (args.Length != 2 && args.Length != 3) throw new Exception("Invalid arguments specified");

					// Parse local (if specified) and server end point
					IPEndPoint bindEP = null;
					if (args.Length == 3) bindEP = TryParseIPEndPoint(args[1], IPAddress.Any);
					var serverEP = TryParseIPEndPoint(args[args.Length - 1], null);
					if (serverEP == null) throw new Exception("Invalid server address specified");

					ClientMain(bindEP, serverEP);
				}
				else if (args[0] == "p") // Puncher server
				{
					if (args.Length != 2) throw new Exception("Invalid arguments specified");

					// Parse local end point
					IPEndPoint bindEP = TryParseIPEndPoint(args[1], IPAddress.Any);
					if (bindEP == null) throw new Exception("Invalid local address specified");

					PuncherServerMain(bindEP);
				}
				else
				{
					throw new Exception("Invalid arguments specified");
				}
			}
			catch (Exception ex)
			{
				Console.Error.WriteLine(ex.Message);
			}
		}

		private static IPEndPoint TryParseIPEndPoint(string s, IPAddress ipa)
		{
			var parts = s.Split(new[] { ':' }, 2);

			// Parse port
			ushort port;
			if (!ushort.TryParse(parts[parts.Length - 1], out port)) return null;
			// Parse IP address if present
			if (parts.Length == 2)
			{
				if (!IPAddress.TryParse(s, out ipa) || ipa.AddressFamily != AddressFamily.InterNetwork)
				{
					ipa = Dns.GetHostEntry(parts[0]).AddressList.First(e => e.AddressFamily == AddressFamily.InterNetwork);
				}
			}
			if (ipa == null) return null;
			
			return new IPEndPoint(ipa, port);
		}

		private static Socket CreateUdpSocket()
		{
			var s = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);

			// Ignore "port unreachable" ICMP message
			const int SIO_UDP_CONNRESET = unchecked((int)0x9800000C);
			s.IOControl(SIO_UDP_CONNRESET, new byte[] { 0, 0, 0, 0 }, null);

			return s;
		}

		private static void ServerMain(IPEndPoint bindEP, IPEndPoint puncherEP)
		{
			var receiveBuffer = new byte[65536];
			using (var s = CreateUdpSocket())
			{
				s.Bind(bindEP);

				// Send punch request
				var punchRequest = CreatePunchRequest(puncherEP);
				if (s.SendTo(punchRequest, puncherEP) != punchRequest.Length)
				{
					throw new Exception("Could not send punch request");
				}

				while (true)
				{
					// Receive packet
					EndPoint remoteEP = new IPEndPoint(IPAddress.Any, 0);
					var numReceived = s.ReceiveFrom(receiveBuffer, ref remoteEP);

					// Got response from puncher server?
					var externalAddress = ReadPunchResponse(receiveBuffer, numReceived);
					if (externalAddress != null)
					{
						Console.WriteLine($"Puncher response: External address is {externalAddress}");
						continue;
					}

					// Echo if packet is not a puncher response
					Console.WriteLine($"Received {numReceived} byte(s) from {remoteEP}");
					if (s.SendTo(receiveBuffer, numReceived, SocketFlags.None, remoteEP) != numReceived)
					{
						throw new Exception("Could not send echo response");
					}
				}
			}
		}

		private static void ClientMain(IPEndPoint bindEP, IPEndPoint serverEP)
		{
			var buffer = new byte[1];

			using (var s = CreateUdpSocket())
			{
				if (bindEP != null) s.Bind(bindEP);
				s.Connect(serverEP);

				// Send and receive packet
				var numSent = s.Send(buffer);
				Console.WriteLine($"Sent {numSent} byte(s)");
				EndPoint remoteEP = new IPEndPoint(IPAddress.Any, 0);
				var numReceived = s.ReceiveFrom(buffer, ref remoteEP);
				Console.WriteLine($"Received {numReceived} byte(s) from {remoteEP}");
			}
		}

		private static void PuncherServerMain(IPEndPoint bindEP)
		{
			var receiveBuffer = new byte[65536];

			using (var s = CreateUdpSocket())
			{
				s.Bind(bindEP);

				while (true)
				{
					// Receive packet
					EndPoint remoteEP = new IPEndPoint(IPAddress.Any, 0);
					var numReceived = s.ReceiveFrom(receiveBuffer, ref remoteEP);

					// Send response if this is a punch request
					if (IsPunchRequest(receiveBuffer, numReceived))
					{
						var punchResponse = CreatePunchResponse((IPEndPoint)remoteEP);
						if (s.SendTo(punchResponse, remoteEP) != punchResponse.Length)
						{
							throw new Exception("Could not send punch response");
						}

						Console.WriteLine($"Punched back at {remoteEP}");
                    }
				}
			}
		}

		private static byte[] CreatePunchRequest(IPEndPoint remoteEP)
		{
			var request = new byte[ConnPacketLen];
			using (var bw = new BinaryWriter(new MemoryStream(request, true)))
			{
				bw.Write((byte)IPTypeID.IPID_Conn); // Status byte
				bw.Write(0u); // Packet number
				bw.Write(C4NetIOUDP__iVersion); // Protocol version
				Write_sockaddr_in(bw, remoteEP); // Address
			}
			return request;
		}

		private static bool IsPunchRequest(byte[] request, int size)
		{
			if (size != ConnPacketLen) return false;

			using (var br = new BinaryReader(new MemoryStream(request)))
			{
				if (br.ReadByte() != (byte)IPTypeID.IPID_Conn) return false; // Status byte
				br.ReadUInt32(); // Packet number
				if (br.ReadUInt32() != C4NetIOUDP__iVersion) return false; // Protocol version
				return true;
			}
		}

		private static byte[] CreatePunchResponse(IPEndPoint externalEP)
		{
			var response = new byte[ConnOKPacketLen];
			using (var bw = new BinaryWriter(new MemoryStream(response, true)))
			{
				bw.Write((byte)IPTypeID.IPID_ConnOK); // Status byte
				bw.Write(0u); // Packet number
				bw.Write((int)MCMode.MCM_NoMC); // MCMode
				Write_sockaddr_in(bw, externalEP); // Address
			}
			return response;
		}

		private static IPEndPoint ReadPunchResponse(byte[] response, int size)
		{
			if (size != ConnOKPacketLen) return null;

			using (var br = new BinaryReader(new MemoryStream(response)))
			{
				if (br.ReadByte() != (byte)IPTypeID.IPID_ConnOK) return null; // Status byte
				br.ReadUInt32(); // Packet number
				br.ReadInt32(); // MCMode
				return Read_sockaddr_in(br);
			}
		}

		private static IPEndPoint Read_sockaddr_in(BinaryReader br)
		{
			if (br.ReadUInt16() != (ushort)AddressFamily.InterNetwork) return null; // sin_family
			var port = (ushort)IPAddress.NetworkToHostOrder(br.ReadInt16()); // sin_port
			var ipa = new IPAddress(br.ReadBytes(4)); // sin_addr
			return new IPEndPoint(ipa, port);
		}

		private static void Write_sockaddr_in(BinaryWriter bw, IPEndPoint ep)
		{
			bw.Write((ushort)AddressFamily.InterNetwork); // sin_family
			bw.Write(IPAddress.HostToNetworkOrder((short)ep.Port)); // sin_port
			bw.Write(ep.Address.GetAddressBytes()); // sin_addr
		}
	}
}
